/*
  ==============================================================================

  RoundTripLatency test process to measure Round-Trip Latency (RTL) using Juce
  Copyright 2023 Mathieu Pavageau https://repetito.com
 
  input:  ________________________________|_____________________________________
  output: ______________________________|_______________________________________
                                        RTL
 
  The RoundTripeLatency::run() process will open the default device and play a digital
  click in the middle of 2 seconds of silence.

  While doing so it will fill a stereo buffer with audio from the first input
  channel on the left, and audio from the first output channel on the right.
 
  If the output volume of the the audio device is high enough, and if the input
  channel is connected to a mic that can pick up the output, the input audio should
  record the click from the output audio.
 
  If the audio device volume was high enough and if no noisy input was recorded
  at the same time, finding the max sample in the recorded input audio measures
  the In/Out latency. This information is logged.

  The 2 second stereo buffer with input on left and output on right is saved on
  the desktop for eventual further analysis. The filename gives the above measured
  info.
 
 ==============================================================================
*/

#define AUDIO_DEVICE_INDEX 0 // if negative, app uses default device, with default settings
#define AUDIO_DEVICE_BUFFER_SIZE 512 // must be one of the available buffer sizes
#define AUDIO_DEVICE_SAMPLING_RATE 48000 // must be one of the available sampling rates
#define CREATE_WAVE_FILE 0 // many tests will create many files on desktop..


#include <atomic>
#include <chrono>
#include <memory>
#include <thread>

#include "JuceIncludes.h"

#include "RoundTripLatency.h"

namespace RoundTripLatency {

constexpr int durationSeconds = 2; // 2 seconds of audio
constexpr int channelCount = 2; // first input channel on left, first output channel on right
constexpr int digitalAudioClickSampleDuration = (AUDIO_DEVICE_SAMPLING_RATE > 48000) ? 8 : 4; // to have similar audio as 44.1/48 kHz when using 88.2/96 kHz
constexpr int inputChannelWithTheMic = 0; // change this to use other input channels than channel 0

#if JUCE_MAJOR_VERSION == 7 && JUCE_BUILDNUMBER > 2
#define JUCE_USES_NEW_AUDIO_DEVICE_IO_CALLBACK 1
#else
#define JUCE_USES_NEW_AUDIO_DEVICE_IO_CALLBACK 0
#endif

struct IOBuffers : juce::AudioIODeviceCallback
{
#if JUCE_USES_NEW_AUDIO_DEVICE_IO_CALLBACK
    void audioDeviceIOCallbackWithContext(const float* const* inputChannelData,
        int numInputChannels,
        float* const* outputChannelData,
        int numOutputChannels,
        int numSamples,
        const juce::AudioIODeviceCallbackContext& /*context*/) override
#else
    void audioDeviceIOCallback(const float** inputChannelData,
        int numInputChannels,
        float** outputChannelData,
        int numOutputChannels,
        int numSamples) override
#endif
    {
        m_bufferSampleCount = numSamples;

        // zero outputs
        for (int outputChannel = 0; outputChannel < numInputChannels; ++outputChannel)
        {
            std::memset(outputChannelData[outputChannel], 0, numSamples * sizeof(float));
        }

        // "play" a digital click lasting digitalAudioClickSampleDuration samples
        // after one second of audio - ie m_samplingRate samples
        if (m_nextSamplePosToWriteTo <= m_samplingRate + digitalAudioClickSampleDuration && m_nextSamplePosToWriteTo + numSamples + digitalAudioClickSampleDuration > m_samplingRate)
        {
            for (int click = 0; click < digitalAudioClickSampleDuration; ++click)
            {
                auto offset = click + m_samplingRate - m_nextSamplePosToWriteTo;
                if (offset >= 0 && offset < numSamples)
                {
                    for (int outputChannel = 0; outputChannel < numInputChannels; ++outputChannel)
                    {
                        outputChannelData[outputChannel][offset] = 0.5f; // sounds bad for sure
                    }
                }
            }
        }

        // fill internal buffer
        if (m_nextSamplePosToWriteTo < m_totalSampleCount)
        {
            const int sizeCanDo = std::min(m_totalSampleCount - m_nextSamplePosToWriteTo, numSamples);

            // copy inputChannelWithTheMic to left
            m_buffers.copyFrom(0, m_nextSamplePosToWriteTo, inputChannelData[inputChannelWithTheMic], sizeCanDo);

            // copy output 0 to right (all outputs have the click)
            m_buffers.copyFrom(1, m_nextSamplePosToWriteTo, outputChannelData[0], sizeCanDo);

            m_nextSamplePosToWriteTo += sizeCanDo;

            if (m_nextSamplePosToWriteTo == m_totalSampleCount)
            {
                // we're finished!
                m_done.store(true);
            }
        }
    }

    void audioDeviceAboutToStart(juce::AudioIODevice* device) override
    {
        // reset internals and store device info
        m_samplingRate = static_cast<int>(device->getCurrentSampleRate());
        m_nextSamplePosToWriteTo = 0;
        m_totalSampleCount = durationSeconds * m_samplingRate;
        m_deviceName = device->getName();
        
        m_inputLatency = device->getInputLatencyInSamples();
        m_outputLatency = device->getOutputLatencyInSamples();

        m_done.store(false);
 
        m_buffers.setSize(channelCount, m_totalSampleCount);
    }

    void audioDeviceStopped() override {}

    // use device name and device info to create unique juce::File on desktop
    juce::File getWaveFile() const
    {
        const juce::String extension(".wav");
        juce::String name(juce::File::getSpecialLocation(juce::File::userDesktopDirectory).getFullPathName() + juce::File::getSeparatorChar() + juce::String("RoundTripLatency_") + juce::SystemStats::getJUCEVersion() + juce::String("_") + m_deviceName + juce::String("_") + juce::String(m_bufferSampleCount) + juce::String("_@") + juce::String(m_samplingRate) + juce::String("_Hz_RTL_") + juce::String(getIoLatency().m_milliseconds, 1) + juce::String("_ms"));

        auto file = juce::File(name + extension);
        if (file.exists() == true)
        {
            // we need to add an index..
            int index = 1;

            const juce::String underscore("_");

            while (1)
            {
                file = juce::File(name + underscore + juce::String(index) + extension);
                if (file.exists() == false)
                {
                    return file;
                }

                ++index;
            }
        }
        else
        {
            return file;
        }
    }

    struct IOLatency
    {
        int m_sampleCount;
        float m_milliseconds;
    };
    IOLatency getIoLatency() const
    {
        // find max in input signal
        int maxSamplePos = 0;
        float maxSampleValue = 0.f;
        
        for (int sample = 0; sample < m_nextSamplePosToWriteTo; ++sample)
        {
            // input is on the left
            const auto absValue = std::abs(m_buffers.getArrayOfReadPointers()[0][sample]);
            if (absValue > maxSampleValue)
            {
                maxSampleValue = absValue;
                maxSamplePos = sample;
            }
        }

        // pos should be greater than m_samplingRate, since we play the digital click at one second
        // io latency is the difference between when we recorded the click and when we played it
        // (open the generated wavefiles if this is not clear)
        if (maxSamplePos > m_samplingRate)
        {
            const auto ioLatencyInSamples = maxSamplePos - m_samplingRate;
            const auto ioLatencyInMilliseconds = 1000.f * static_cast<float>(ioLatencyInSamples) / static_cast<float>(m_samplingRate);
            
            return { ioLatencyInSamples, ioLatencyInMilliseconds };
        }
            
        return { 0, 0.f };
    }

    juce::AudioSampleBuffer m_buffers;

    int m_nextSamplePosToWriteTo{ 0 };
    int m_totalSampleCount{ 0 };
    int m_samplingRate{ 0 };
    int m_bufferSampleCount{ 0 };
    int m_inputLatency{ 0 };
    int m_outputLatency{ 0 };
    juce::String m_deviceName;

    std::atomic<bool> m_done{ false };
};

void run()
{
    juce::AudioDeviceManager deviceManager;

    IOBuffers buffers;
    
    bool useDefaultDevice = true;
    juce::AudioDeviceManager::AudioDeviceSetup setup;

    auto& deviceTypes = deviceManager.getAvailableDeviceTypes();
    for (auto& deviceType : deviceTypes)
    {
        deviceType->scanForDevices();
        
        const auto inputDeviceNames = deviceType->getDeviceNames(true);
        const auto outputDeviceNames = deviceType->getDeviceNames(false);

        jassert(inputDeviceNames.size() == outputDeviceNames.size());
        for (int deviceIndex = 0; deviceIndex < inputDeviceNames.size(); ++deviceIndex)
        {
            juce::Logger::writeToLog(juce::String("Device ") + juce::String(deviceIndex) + juce::String(": input name: ") + inputDeviceNames[deviceIndex] + juce::String(" - output name: ") + outputDeviceNames[deviceIndex]);
            
            if (deviceIndex == AUDIO_DEVICE_INDEX)
            {
                setup.inputDeviceName = inputDeviceNames[deviceIndex];
                setup.outputDeviceName = outputDeviceNames[deviceIndex];
                
                useDefaultDevice = false;
            }
        }
    }

    deviceManager.addAudioCallback(&buffers);

    if (useDefaultDevice)
    {
        deviceManager.initialiseWithDefaultDevices(2, 2);
    }
    else
    {
        setup.useDefaultInputChannels = true;
        setup.useDefaultOutputChannels = true;
        setup.sampleRate = AUDIO_DEVICE_SAMPLING_RATE;
        setup.bufferSize = AUDIO_DEVICE_BUFFER_SIZE;

        deviceManager.initialise(2, 2, nullptr, false, juce::String(), &setup);
    }

    while (buffers.m_done.load() == false)
    {
        std::this_thread::sleep_for(std::chrono::milliseconds(50));
    }

    if (buffers.m_nextSamplePosToWriteTo > 0)
    {
        // save stereo 2 second buffer in wavefile on desktop
        auto file = buffers.getWaveFile();

#if CREATE_WAVE_FILE
#if JUCE_MAJOR_VERSION > 5
        std::unique_ptr<juce::FileOutputStream> fileStream = file.createOutputStream();
#else
        juce::FileOutputStream* fileStream = file.createOutputStream();
#endif

        juce::WavAudioFormat wavFormat;
#if JUCE_MAJOR_VERSION > 5
        if (auto writer = std::unique_ptr<juce::AudioFormatWriter>(wavFormat.createWriterFor(fileStream.release(), buffers.m_samplingRate, 2, 16, {}, 0)))
#else
        if (auto writer = std::unique_ptr<juce::AudioFormatWriter>(wavFormat.createWriterFor(fileStream, buffers.m_samplingRate, 2, 16, {}, 0)))
#endif
        {
            writer->writeFromAudioSampleBuffer(buffers.m_buffers, 0, buffers.m_nextSamplePosToWriteTo);
        }
#endif // CREATE_WAVE_FILE

        // log info for debug
        const auto ioLatency = buffers.getIoLatency();
        juce::Logger::writeToLog(juce::String("\nRound Trip Latency results: "));
        juce::Logger::writeToLog(juce::String("Sampling rate: ") + juce::String(buffers.m_samplingRate) + juce::String(" Hz - buffer sample size: ") + juce::String(buffers.m_bufferSampleCount));
        juce::Logger::writeToLog(juce::String("Device gives: Input Latency: ") + juce::String(buffers.m_inputLatency) + juce::String(" samples - Output Latency: ") + juce::String(buffers.m_outputLatency) + juce::String(" samples (sum is ") + juce::String(buffers.m_inputLatency + buffers.m_outputLatency) + juce::String(")"));
        juce::Logger::writeToLog(juce::String("Measured Round-Trip Latency (RTL) is ") + juce::String(ioLatency.m_sampleCount) + juce::String(" samples = ") + juce::String(ioLatency.m_milliseconds, 1) + juce::String(" ms"));
    }
}

} // namespace RoundTripLatency
